package meerkat.Registry;

import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.VoterRegistry;

import java.text.ParseException;
import java.util.*;

/**
 * TODO: add logging to this utils
 */

/**
 * Created by Vladimir Eliezer Tokarev  on 1/15/2016.
 * adds extra functionality to Messages collections
 */
public abstract class CollectionMessagesUtils {

    public static class EmptyListException extends Exception {
        public EmptyListException(String message) {
            super(message);
        }
    }

    /**
     * Converts lost of UnsignedBulletinBoardMessage to VoterRegistryMessages
     * @param messages list<VoterRegistryMessage>
     * @return List<VoterRegistryMessage>
     */
   public static List<UnsignedBulletinBoardMessage> ConvertToVoterRegistryMessages(List<BulletinBoardMessage> messages){
       List<UnsignedBulletinBoardMessage> voterMessages = new ArrayList<>();
       for (int i = 0 ; i < messages.size() ; i++)
       {
           voterMessages.add(messages.get(i).getMsg());
       }
       return voterMessages;
   }

    /**
     * Return true if message1 was created before message 2
     * @param message1 BulletinBoardMessage object
     * @param message2 BulletinBoardMessage object
     * @return  boolean
     * @throws InvalidProtocolBufferException
     */
    private static boolean FirstBeforeSecond(BulletinBoardMessage message1, BulletinBoardMessage message2)
            throws InvalidProtocolBufferException {
        try {
            long firstCreationTime =  VoterRegistry.VoterRegistryMessage.newBuilder()
                    .mergeFrom(message1.getMsg().getData()).build().getCreationMiliseconds();
            long secondCreationTime = VoterRegistry.VoterRegistryMessage.newBuilder()
                    .mergeFrom(message2.getMsg().getData()).build().getCreationMiliseconds();
            return firstCreationTime < secondCreationTime;
        } catch (InvalidProtocolBufferException e) {
            e.printStackTrace();
            throw e;
        }
    }

    /**
     * Gets map of GroupId to basicMessage, where the basicMessages are the last actions for those groups
     * @param messages List<VoterRegistryMessage>
     * @return     Map{String:VoterRegistryMessage}
     * @throws ParseException
     */
    public static Map<String, BulletinBoardMessage> GetLatestGroupsActions(List<BulletinBoardMessage> messages)
            throws ParseException, InvalidProtocolBufferException {
        Map<String, BulletinBoardMessage> groupIdToMessage = new HashMap<>(messages.size());

        // iterate trough all the messages and put into the map the last updated groups actions
        for (int i = 0 ; i < messages.size() ; i++) {
            BulletinBoardMessage message = messages.get(i);
            String groupId = GetTagByName(message.getMsg().getTagList(), RegistryTags.GROUP_ID_TAG);
            BulletinBoardMessage temp = groupIdToMessage.get(groupId);

            if (temp != null && temp != message) {
                if (FirstBeforeSecond(temp, message)) {
                    groupIdToMessage.put(groupId, message);
                }
            }
            groupIdToMessage.put(groupId, message);
        }
        return groupIdToMessage;
    }

    /**
     * Gets list of groups ids of the basicMessages that carried the adding to group tag
     * @param groupIdToMessage Map<String, VoterRegistryMessage>
     * @return List<String>
     */
     public static List<String> GetListOfGroupIds(Map<String, BulletinBoardMessage>  groupIdToMessage) {
         List<String> groupsIds = new ArrayList<>(groupIdToMessage.size());
         Iterator entries = groupIdToMessage.entrySet().iterator();
         while (entries.hasNext()) {
             Map.Entry tuple = (Map.Entry) entries.next();
             BulletinBoardMessage message = (BulletinBoardMessage) tuple.getValue();
             String groupId = GetTagByName(message.getMsg().getTagList(), RegistryTags.GROUP_ID_TAG);
             groupId.replace(RegistryTags.GROUP_ID_TAG, "");
             groupsIds.add(groupId);
         }
         return groupsIds;
     }

    /**
     * Gets the message with the latest timestamp from messages
     * @param messages  List<VoterRegistryMessage>
     * @return VoterRegistryMessage
     * @throws ParseException
     * @throws EmptyListException
     */
    public static BulletinBoardMessage GetLatestMessage(List<BulletinBoardMessage> messages)
            throws ParseException, EmptyListException, InvalidProtocolBufferException {
        if (messages.size() == 0 ){
            throw new EmptyListException("The list of messages passed to GetLatestMessage is empty.");
        }
        BulletinBoardMessage LatestMessage = messages.get(0);
        for (int i = 0 ; i < messages.size() ; i++) {
            BulletinBoardMessage message = messages.get(i);
            if (FirstBeforeSecond(message, LatestMessage)) {
                LatestMessage = message;
            }
        }
        return LatestMessage;
    }

    /**
     * Gets the wanted tag from given basic message
     *
     * @param content the name of the tag
     * @param tags the list of tags
     * @return string
     */
    public static String GetTagByName(List<String> tags, String content) {
        for (int i = 0 ; i < tags.size() ; i++) {
            String tag = tags.get(i);
            if (tag.startsWith(content) || tag.endsWith(content)) {
                return tag;
            }
        }
        return null;
    }


    /**
     * Gets messages that have the wanted id tag and have or the ADD_TO_GROUP_TAG or REMOVE_FROM_GROUP_TAG
     *
     * @param tags the tags based on which the messages will be filtered
     * @return MessageFilterList.
     */
    public static MessageFilterList GenerateFiltersFromTags(List<String> tags) {
        MessageFilterList.Builder filters = MessageFilterList.newBuilder();

        if (tags.isEmpty()){
            return filters.build();
        }

        for (int i = 0 ;i < tags.size() ; i++) {
            String tag = tags.get(i);
            MessageFilter.Builder filter = MessageFilter.newBuilder().setTag(tag).setType(FilterType.TAG);
            filters.addFilter(filter);
        }
        return filters.build();
    }
}
